@@ -47,7 +47,7 @@ static void add_noise(const mjModel* m, mjData* d, mjtStage stage) {
 
       // real or positive: add noise directly, with clamp for positive
       if (m->sensor_datatype[i]==mjDATATYPE_REAL ||
-          m->sensor_datatype[i]==mjDATATYPE_POSITIVE)
+          m->sensor_datatype[i]==mjDATATYPE_POSITIVE) {
         for (int j=0; j<dim; j++) {
           // get random numbers; use only the first one
           rnd[0] = mju_standardNormal(rnd+1);
@@ -65,6 +65,7 @@ static void add_noise(const mjModel* m, mjData* d, mjtStage stage) {
             d->sensordata[adr+j] += rnd[0]*noise;
           }
         }
+      }
 
       // axis or quat: rotate around random axis by random angle
       else {
@@ -72,7 +73,7 @@ static void add_noise(const mjModel* m, mjData* d, mjtStage stage) {
         rnd[0] = mju_standardNormal(rnd+1);
         rnd[2] = mju_standardNormal(rnd+3);
 
-        // scale angle, normalize axis, make quaterion
+        // scale angle, normalize axis, make quaternion
         rnd[0] *= noise;
         mju_normalize3(rnd+1);
         mju_axisAngle2Quat(quat, rnd+1, rnd[0]);
@@ -113,17 +114,17 @@ static void apply_cutoff(const mjModel* m, mjData* d, mjtStage stage) {
       mjtNum cutoff = m->sensor_cutoff[i];
 
       // process all dimensions
-      for (int j=0; j<dim; j++)
-
+      for (int j=0; j<dim; j++) {
         // real: apply on both sides
-        if (m->sensor_datatype[i]==mjDATATYPE_REAL)
-          d->sensordata[adr+j] =
-            mju_min(cutoff, mju_max(-cutoff, d->sensordata[adr+j]));
+        if (m->sensor_datatype[i]==mjDATATYPE_REAL) {
+          d->sensordata[adr+j] = mju_clip(d->sensordata[adr+j], -cutoff, cutoff);
+        }
 
         // positive: apply on positive side only
-        else if (m->sensor_datatype[i]==mjDATATYPE_POSITIVE)
-          d->sensordata[adr+j] =
-            mju_min(cutoff, d->sensordata[adr+j]);
+        else if (m->sensor_datatype[i]==mjDATATYPE_POSITIVE) {
+          d->sensordata[adr+j] = mju_min(cutoff, d->sensordata[adr+j]);
+        }
+      }
     }
   }
 }
@@ -192,6 +193,11 @@ void mj_sensorPos(const mjModel* m, mjData* d) {
   int ne = d->ne, nf = d->nf, nefc = d->nefc;
   mjtNum rvec[3], *xpos, *xmat, *xpos_ref, *xmat_ref;
 
+  // disabled sensors: return
+  if (mjDISABLED(mjDSBL_SENSOR)) {
+    return;
+  }
+
   // process sensors matching stage
   for (int i=0; i<m->nsensor; i++) {
     if (m->sensor_needstage[i]==mjSTAGE_POS) {
@@ -312,6 +318,10 @@ void mj_sensorPos(const mjModel* m, mjData* d) {
         mju_copy3(d->sensordata+adr, d->subtree_com+3*objid);
         break;
 
+      case mjSENS_CLOCK:                                  // clock
+        d->sensordata[adr] = d->time;
+        break;
+
       case mjSENS_USER:                                   // user
         nusersensor++;
         break;
@@ -344,6 +354,11 @@ void mj_sensorVel(const mjModel* m, mjData* d) {
   int ne = d->ne, nf = d->nf, nefc = d->nefc;
   mjtNum xvel[6];
 
+  // disabled sensors: return
+  if (mjDISABLED(mjDSBL_SENSOR)) {
+    return;
+  }
+
   // process sensors matching stage
   int subtreeVel = 0;
   for (int i=0; i<m->nsensor; i++) {
@@ -497,6 +512,11 @@ void mj_sensorAcc(const mjModel* m, mjData* d) {
   mjtNum tmp[6], conforce[6], conray[3];
   mjContact* con;
 
+  // disabled sensors: return
+  if (mjDISABLED(mjDSBL_SENSOR)) {
+    return;
+  }
+
   // process sensors matching stage
   int rnePost = 0;
   for (int i=0; i<m->nsensor; i++) {
