@@ -35,8 +35,8 @@
   #endif
 #endif
 
-#define mjVERSION 220
-#define mjVERSIONSTRING "2.2.0"
+#define mjVERSION 222
+#define mjVERSIONSTRING "2.2.2"
 
 // names of disable flags
 const char* mjDISABLESTRING[mjNDISABLE] = {
@@ -51,7 +51,8 @@ const char* mjDISABLESTRING[mjNDISABLE] = {
   "Warmstart",
   "Filterparent",
   "Actuation",
-  "Refsafe"
+  "Refsafe",
+  "Sensor"
 };
 
 
@@ -154,6 +155,35 @@ void mj_jacBodyCom(const mjModel* m, const mjData* d, mjtNum* jacp, mjtNum* jacr
 
 
 
+// compute subtree-com Jacobian
+void mj_jacSubtreeCom(const mjModel* m, mjData* d, mjtNum* jacp, int body) {
+  int nv = m->nv;
+  mjMARKSTACK;
+  mjtNum* jacp_b = mj_stackAlloc(d, 3*nv);
+
+  // clear output
+  mju_zero(jacp, 3*nv);
+
+  // forward pass starting from body
+  for (int b=body; b<m->nbody; b++) {
+    // end of body subtree, break from the loop
+    if (b > body && m->body_parentid[b] < body) {
+      break;
+    }
+
+    // b is in the body subtree, add mass-weighted Jacobian into jacp
+    mj_jac(m, d, jacp_b, NULL, d->xipos+3*b, b);
+    mju_addToScl(jacp, jacp_b, m->body_mass[b], 3*nv);
+  }
+
+  // normalize by subtree mass
+  mju_scl(jacp, jacp, 1/m->body_subtreemass[body], 3*nv);
+
+  mjFREESTACK;
+}
+
+
+
 // compute geom Jacobian
 void mj_jacGeom(const mjModel* m, const mjData* d, mjtNum* jacp, mjtNum* jacr, int geom) {
   mj_jac(m, d, jacp, jacr, d->geom_xpos + 3*geom, m->geom_bodyid[geom]);
@@ -1199,7 +1229,7 @@ void mj_integratePos(const mjModel* m, mjtNum* qpos, const mjtNum* qvel, mjtNum
 
 
 
-// normalize all quaterions in qpos-type vector
+// normalize all quaternions in qpos-type vector
 void mj_normalizeQuat(const mjModel* m, mjtNum* qpos) {
   // find quaternion fields and normalize
   for (int i=0; i<m->njnt; i++) {
