@@ -32,7 +32,9 @@
 
 //----------------------------- utility functions and macros ---------------------------------------
 
-static const mjtNum IDENTITY[9] = {1,0,0, 0,1,0, 0,0,1};
+static const mjtNum IDENTITY[9] = {1, 0, 0,
+                                   0, 1, 0,
+                                   0, 0, 1};
 
 
 // copy float array
@@ -64,16 +66,18 @@ static void makeLabel(const mjModel* m, mjtObj type, int id, char* label) {
 
 
 // return if there is no space in buffer
-#define START                                                           \
-    if( scn->ngeom>=scn->maxgeom )                                      \
-         { mj_warning(d, mjWARN_VGEOMFULL, scn->maxgeom);               \
-           return; }                                                    \
-    else { thisgeom = scn->geoms + scn->ngeom;                          \
-           mjv_initGeom(thisgeom, mjGEOM_NONE, NULL, NULL, NULL, NULL); \
-           thisgeom->objtype = objtype;                                 \
-           thisgeom->objid = i;                                         \
-           thisgeom->category = category;                               \
-           thisgeom->segid = scn->ngeom;    }
+#define START                                                      \
+    if ( scn->ngeom>=scn->maxgeom ) {                              \
+      mj_warning(d, mjWARN_VGEOMFULL, scn->maxgeom);               \
+      return;                                                      \
+    } else {                                                       \
+      thisgeom = scn->geoms + scn->ngeom;                          \
+      mjv_initGeom(thisgeom, mjGEOM_NONE, NULL, NULL, NULL, NULL); \
+      thisgeom->objtype = objtype;                                 \
+      thisgeom->objid = i;                                         \
+      thisgeom->category = category;                               \
+      thisgeom->segid = scn->ngeom;                                \
+    }
 
 
 // advance counter
@@ -327,29 +331,30 @@ void mjv_initGeom(mjvGeom* geom, int type, const mjtNum* size,
   geom->type = type;
 
   // set size (for XYZ scaling)
-  if (size)
+  if (size) {
     switch (type) {
-    case mjGEOM_SPHERE:
-      geom->size[0] = (float)size[0];
-      geom->size[1] = (float)size[0];
-      geom->size[2] = (float)size[0];
-      break;
+      case mjGEOM_SPHERE:
+        geom->size[0] = (float)size[0];
+        geom->size[1] = (float)size[0];
+        geom->size[2] = (float)size[0];
+        break;
 
-    case mjGEOM_CAPSULE:
-      geom->size[0] = (float)size[0];
-      geom->size[1] = (float)size[0];
-      geom->size[2] = (float)size[1];
-      break;
+      case mjGEOM_CAPSULE:
+        geom->size[0] = (float)size[0];
+        geom->size[1] = (float)size[0];
+        geom->size[2] = (float)size[1];
+        break;
 
-    case mjGEOM_CYLINDER:
-      geom->size[0] = (float)size[0];
-      geom->size[1] = (float)size[0];
-      geom->size[2] = (float)size[1];
-      break;
+      case mjGEOM_CYLINDER:
+        geom->size[0] = (float)size[0];
+        geom->size[1] = (float)size[0];
+        geom->size[2] = (float)size[1];
+        break;
 
-    default:
-      mju_n2f(geom->size, size, 3);
-    } else {
+      default:
+        mju_n2f(geom->size, size, 3);
+    }
+  } else {
     geom->size[0] = 0.1f;
     geom->size[1] = 0.1f;
     geom->size[2] = 0.1f;
@@ -410,8 +415,6 @@ void mjv_initGeom(mjvGeom* geom, int type, const mjtNum* size,
 static void markselected(const mjVisual* vis, mjvGeom* geom) {
   // add emission
   geom->emission += vis->global.glow;
-  // make opaque
-  geom->rgba[3] = 1;
 }
 
 
@@ -443,13 +446,12 @@ static int bodycategory(const mjModel* m, int bodyid) {
   }
 }
 
-
-
 // add abstract geoms
 void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
                   const mjvPerturb* pert, int catmask, mjvScene* scn) {
   int objtype, category;
   mjtNum sz[3], mat[9], selpos[3];
+  mjtNum catenary[3*mjNCATENARY];
   mjtNum *cur, *nxt, *xpos, *xfrc;
   mjtNum vec[3], end[3], axis[3], rod, len, det, tmp[9], quat[4];
   mjtByte broken;
@@ -774,23 +776,26 @@ void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
       if (vopt->actuatorgroup[mjMAX(0, mjMIN(mjNGROUP-1, m->actuator_group[i]))]) {
         // determine extended range
         mjtNum rng[3] = {-1, 0, +1};
+        mjtNum rmin = -1, rmax = 1, act;
         if (m->actuator_ctrllimited[i]) {
-          mjtNum rmin = m->actuator_ctrlrange[2*i];
-          mjtNum rmax = m->actuator_ctrlrange[2*i+1];
-
-          if (rmin>=0) {
-            rng[0] = -1;
-            rng[1] = rmin;
-            rng[2] = rmax;
-          } else if (rmax<=0) {
-            rng[0] = rmin;
-            rng[1] = rmax;
-            rng[2] = +1;
-          } else {
-            rng[0] = rmin;
-            rng[1] = 0;
-            rng[2] = rmax;
-          }
+          rmin = m->actuator_ctrlrange[2*i];
+          rmax = m->actuator_ctrlrange[2*i+1];
+        } else if (vopt->flags[mjVIS_ACTIVATION] && m->actuator_actlimited[i]) {
+          rmin = m->actuator_actrange[2*i];
+          rmax = m->actuator_actrange[2*i+1];
+        }
+        if (rmin>=0) {
+          rng[0] = -1;
+          rng[1] = rmin;
+          rng[2] = rmax;
+        } else if (rmax<=0) {
+          rng[0] = rmin;
+          rng[1] = rmax;
+          rng[2] = +1;
+        } else {
+          rng[0] = rmin;
+          rng[1] = 0;
+          rng[2] = rmax;
         }
 
         // adjust small ranges
@@ -801,8 +806,12 @@ void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
           rng[2] = rng[1] + mjMINVAL;
         }
 
-        // clamp ctrl to extended range
-        mjtNum act = mjMIN(rng[2], mjMAX(rng[0], d->ctrl[i]));
+        // clamp act to extended range
+        if (vopt->flags[mjVIS_ACTIVATION] && m->actuator_dyntype[i]) {
+          act = mju_clip(d->act[i-(m->nu-m->na)], rng[0], rng[2]);
+        } else {
+          act = mju_clip(d->ctrl[i], rng[0], rng[2]);
+        }
 
         // compute interpolants
         float amin, amean, amax;
@@ -828,23 +837,49 @@ void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
         int j = m->actuator_trnid[2*i];
 
         // slide and hinge joint actuators
-        if ((m->actuator_trntype[i]==mjTRN_JOINT || m->actuator_trntype[i]==mjTRN_JOINTINPARENT) &&
-            (m->jnt_type[j]==mjJNT_HINGE || m->jnt_type[j]==mjJNT_SLIDE)) {
-          // set length(1) and width(0) of the connectors
-          sz[1] = m->vis.scale.actuatorlength * scl;
-          sz[0] = m->vis.scale.actuatorwidth * scl;
-
+        if (m->actuator_trntype[i]==mjTRN_JOINT ||
+            m->actuator_trntype[i]==mjTRN_JOINTINPARENT ||
+            m->actuator_trntype[i]==mjTRN_SITE) {
           START
 
-          // make geom
-          mjv_makeConnector(thisgeom,
-                            m->jnt_type[j]==mjJNT_SLIDE ? mjGEOM_ARROW : mjGEOM_ARROW1, sz[0],
-                            d->xanchor[3*j+0],
-                            d->xanchor[3*j+1],
-                            d->xanchor[3*j+2],
-                            d->xanchor[3*j+0] + sz[1]*d->xaxis[3*j+0],
-                            d->xanchor[3*j+1] + sz[1]*d->xaxis[3*j+1],
-                            d->xanchor[3*j+2] + sz[1]*d->xaxis[3*j+2]);
+          // site actuators
+          if (m->actuator_trntype[i]==mjTRN_SITE) {
+            // inflate sizes by 5%
+            mju_scl3(sz, m->site_size+3*j, 1.05);
+
+            // make geom
+            mjv_initGeom(thisgeom,
+                         m->site_type[j], sz,
+                         d->site_xpos + 3*j,
+                         d->site_xmat + 9*j,
+                         thisgeom->rgba);
+          } else if (m->jnt_type[j]==mjJNT_HINGE || m->jnt_type[j]==mjJNT_SLIDE) {
+            // set length(1) and width(0) of the connectors
+            sz[1] = m->vis.scale.actuatorlength * scl;
+            sz[0] = m->vis.scale.actuatorwidth * scl;
+
+            // make geom
+            mjv_makeConnector(thisgeom,
+                              m->jnt_type[j]==mjJNT_SLIDE ? mjGEOM_ARROW : mjGEOM_ARROW1, sz[0],
+                              d->xanchor[3*j+0],
+                              d->xanchor[3*j+1],
+                              d->xanchor[3*j+2],
+                              d->xanchor[3*j+0] + sz[1]*d->xaxis[3*j+0],
+                              d->xanchor[3*j+1] + sz[1]*d->xaxis[3*j+1],
+                              d->xanchor[3*j+2] + sz[1]*d->xaxis[3*j+2]);
+          }
+
+          // ball or free joint
+          else if (m->jnt_type[j]==mjJNT_BALL || m->jnt_type[j]==mjJNT_FREE) {
+            sz[0] = sz[1] = sz[2] = m->vis.scale.jointlength * scl * 0.33;
+
+            // make geom
+            mjv_initGeom(thisgeom,
+                         m->jnt_type[j]==mjJNT_BALL ? mjGEOM_SPHERE : mjGEOM_BOX, sz,
+                         d->xanchor + 3*j,
+                         d->xmat + 9*m->jnt_bodyid[j],
+                         thisgeom->rgba);
+          }
 
           // set interpolated color
           f2f(thisgeom->rgba, rgba, 4);
@@ -857,6 +892,34 @@ void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
           FINISH
         }
 
+        // body actuators
+        else if (m->actuator_trntype[i]==mjTRN_BODY) {
+          // iterate over body's geoms
+          int geomnum = m->body_geomnum[j];
+          int geomadr = m->body_geomadr[j];
+          for (int k=geomadr; k<geomadr+geomnum; k++) {
+            int geomtype = m->geom_type[k];
+            // add inflated geom if it is a regular primitive
+            if (geomtype != mjGEOM_PLANE && geomtype != mjGEOM_HFIELD && geomtype != mjGEOM_MESH) {
+              START
+              // inflate sizes by 5%
+              mju_scl3(sz, m->geom_size+3*k, 1.05);
+
+              // make geom
+              mjv_initGeom(thisgeom,
+                           m->geom_type[k], sz,
+                           d->geom_xpos + 3*k,
+                           d->geom_xmat + 9*k,
+                           thisgeom->rgba);
+
+              // set interpolated color
+              f2f(thisgeom->rgba, rgba, 4);
+
+              FINISH
+            }
+          }
+        }
+
         // spatial tendon actuators
         else if (m->actuator_trntype[i]==mjTRN_TENDON && d->ten_wrapnum[j]) {
           for (int k=d->ten_wrapadr[j]; k<d->ten_wrapadr[j]+d->ten_wrapnum[j]-1; k++) {
@@ -1287,27 +1350,77 @@ void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
   if (vopt->flags[mjVIS_TENDON] && (category & catmask)) {
     for (int i=0; i<m->ntendon; i++) {
       if (vopt->tendongroup[mjMAX(0, mjMIN(mjNGROUP-1, m->tendon_group[i]))]) {
-        for (int j=d->ten_wrapadr[i]; j<d->ten_wrapadr[i]+d->ten_wrapnum[i]-1; j++) {
-          if (d->wrap_obj[j]!=-2 && d->wrap_obj[j+1]!=-2) {
+        // conditions for drawing a catenary
+        int draw_catenary =
+            !mjDISABLED(mjDSBL_GRAVITY)           &&  // gravity enabled
+            mju_norm3(m->opt.gravity) > mjMINVAL  &&  // gravity strictly nonzero
+            m->tendon_num[i] == 2                 &&  // only two sites on the tendon
+            m->tendon_limited[i] == 1             &&  // limited length range
+            m->tendon_range[2*i] == 0             &&  // range lower-bound is 0
+            m->tendon_stiffness[i] == 0           &&  // no stiffness
+            m->tendon_damping[i] == 0             &&  // no damping
+            m->tendon_frictionloss[i] == 0;           // no frictionloss
+
+        // conditions not met: draw straight lines
+        if (!draw_catenary) {
+          for (int j=d->ten_wrapadr[i]; j<d->ten_wrapadr[i]+d->ten_wrapnum[i]-1; j++) {
+            if (d->wrap_obj[j]!=-2 && d->wrap_obj[j+1]!=-2) {
+              START
+
+              // determine width: smaller for segments inside wrapping objects
+              if (d->wrap_obj[j]>=0 && d->wrap_obj[j+1]>=0) {
+                sz[0] = 0.5 * m->tendon_width[i];
+              } else {
+                sz[0] = m->tendon_width[i];
+              }
+
+              // construct geom
+              mjv_makeConnector(thisgeom, mjGEOM_CAPSULE, sz[0],
+                                d->wrap_xpos[3*j], d->wrap_xpos[3*j+1], d->wrap_xpos[3*j+2],
+                                d->wrap_xpos[3*j+3], d->wrap_xpos[3*j+4], d->wrap_xpos[3*j+5]);
+
+              // set material if given
+              setMaterial(m, thisgeom, m->tendon_matid[i], m->tendon_rgba+4*i, vopt->flags);
+
+              // vopt->label: only the first segment
+              if (vopt->label==mjLABEL_TENDON && j==d->ten_wrapadr[i]) {
+                makeLabel(m, mjOBJ_TENDON, i, thisgeom->label);
+              }
+
+              FINISH
+            }
+          }
+        }
+
+        // special case handling of string-like tendons under gravity
+        else {
+          // two hanging points: x0, x1
+          mjtNum x0[3], x1[3];
+          mju_copy3(x0, d->wrap_xpos + 3*d->ten_wrapadr[i]);
+          mju_copy3(x1, d->wrap_xpos + 3*d->ten_wrapadr[i] + 3);
+
+          // length of the tendon
+          mjtNum length = m->tendon_range[2*i+1];
+
+          // points along catenary path
+          int npoints = mjv_catenary(x0, x1, m->opt.gravity, length, catenary);
+
+          // draw npoints-1 segments
+          for (int j=0; j<npoints-1; j++) {
             START
 
-            // determine width: smaller for segments inside wrapping objects
-            if (d->wrap_obj[j]>=0 && d->wrap_obj[j+1]>=0) {
-              sz[0] = 0.5 * m->tendon_width[i];
-            } else {
-              sz[0] = m->tendon_width[i];
-            }
+            sz[0] = m->tendon_width[i];
 
             // construct geom
             mjv_makeConnector(thisgeom, mjGEOM_CAPSULE, sz[0],
-                              d->wrap_xpos[3*j], d->wrap_xpos[3*j+1], d->wrap_xpos[3*j+2],
-                              d->wrap_xpos[3*j+3], d->wrap_xpos[3*j+4], d->wrap_xpos[3*j+5]);
+                              catenary[3*j], catenary[3*j+1], catenary[3*j+2],
+                              catenary[3*j+3], catenary[3*j+4], catenary[3*j+5]);
 
             // set material if given
             setMaterial(m, thisgeom, m->tendon_matid[i], m->tendon_rgba+4*i, vopt->flags);
 
             // vopt->label: only the first segment
-            if (vopt->label==mjLABEL_TENDON && j==d->ten_wrapadr[i]) {
+            if (vopt->label==mjLABEL_TENDON && npoints/2) {
               makeLabel(m, mjOBJ_TENDON, i, thisgeom->label);
             }
 
@@ -1321,8 +1434,8 @@ void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
   // slider-crank
   objtype = mjOBJ_ACTUATOR;
   category = mjCAT_DYNAMIC;
-  if ((category & catmask))
-    for (int i=0; i<m->nu; i++)
+  if ((category & catmask)) {
+    for (int i=0; i<m->nu; i++) {
       if (m->actuator_trntype[i]==mjTRN_SLIDERCRANK) {
         // get data
         int j = m->actuator_trnid[2*i];                 // crank
@@ -1370,6 +1483,8 @@ void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
         }
         FINISH
       }
+    }
+  }
 
   // center of mass for root bodies
   objtype = mjOBJ_UNKNOWN;
@@ -1491,33 +1606,31 @@ void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* vopt,
   objtype = mjOBJ_EQUALITY;
   category = mjCAT_DECOR;
   if (vopt->flags[mjVIS_CONSTRAINT] && (category & catmask) && m->neq) {
-    // connect
+    // connect or weld
     for (int i=0; i<m->neq; i++) {
-      if (m->eq_active[i] && m->eq_type[i]==mjEQ_CONNECT) {
+      if (m->eq_active[i] && (m->eq_type[i]==mjEQ_CONNECT || m->eq_type[i]==mjEQ_WELD)) {
         // compute endpoints in global coordinates
-        int j = m->eq_obj1id[i];
-        int k = m->eq_obj2id[i];
-        mju_rotVecMat(vec, m->eq_data+mjNEQDATA*i, d->xmat+9*j);
+        int j = m->eq_obj1id[i], k = m->eq_obj2id[i];
+        mju_rotVecMat(vec, m->eq_data+mjNEQDATA*i+3*(m->eq_type[i]==mjEQ_WELD), d->xmat+9*j);
         mju_addTo3(vec, d->xpos+3*j);
-        mju_rotVecMat(end, m->eq_data+mjNEQDATA*i+3, d->xmat+9*k);
+        mju_rotVecMat(end, m->eq_data+mjNEQDATA*i+3*(m->eq_type[i]==mjEQ_CONNECT), d->xmat+9*k);
         mju_addTo3(end, d->xpos+3*k);
 
-        // connect endpoints
-        START
-
         // construct geom
         sz[0] = scl * m->vis.scale.constraint;
-        mjv_makeConnector(thisgeom, mjGEOM_CAPSULE, sz[0],
-                          vec[0], vec[1], vec[2],
-                          end[0], end[1], end[2]);
 
-        f2f(thisgeom->rgba, m->vis.rgba.constraint, 4);
-
-        // label flag
+        START
+        mjv_initGeom(thisgeom, mjGEOM_SPHERE, sz, vec, d->xmat+9*j, m->vis.rgba.connect);
         if (vopt->label==mjLABEL_CONSTRAINT) {
           makeLabel(m, mjOBJ_EQUALITY, i, thisgeom->label);
         }
+        FINISH
 
+        START
+        mjv_initGeom(thisgeom, mjGEOM_SPHERE, sz, end, d->xmat+9*k, m->vis.rgba.constraint);
+        if (vopt->label==mjLABEL_CONSTRAINT) {
+          makeLabel(m, mjOBJ_EQUALITY, i, thisgeom->label);
+        }
         FINISH
       }
     }
@@ -1733,9 +1846,17 @@ void mjv_updateCamera(const mjModel* m, mjData* d, mjvCamera* cam, mjvScene* scn
 }
 
 
-
-// update skins only
+// update all skins, here for backward API compatibility
 void mjv_updateSkin(const mjModel* m, mjData* d, mjvScene* scn) {
+  mjvOption opt;
+  mjv_defaultOption(&opt);
+  mjv_updateActiveSkin(m, d, scn, &opt);
+  mju_warning("mjv_updateSkin is deprecated, please use mjv_updateActiveSkin.");
+}
+
+
+// update visible skins only
+void mjv_updateActiveSkin(const mjModel* m, mjData* d, mjvScene* scn, const mjvOption* opt) {
   // process skins
   for (int i=0; i<m->nskin; i++) {
     // get info
@@ -1748,111 +1869,113 @@ void mjv_updateSkin(const mjModel* m, mjData* d, mjvScene* scn) {
     memset(scn->skinvert + 3*vertadr, 0, 3*vertnum*sizeof(float));
     memset(scn->skinnormal + 3*vertadr, 0, 3*vertnum*sizeof(float));
 
-    // accumulate positions from all bones
-    for (int j=m->skin_boneadr[i];
-         j<m->skin_boneadr[i]+m->skin_bonenum[i];
-         j++) {
-      // get bind pose
-      mjtNum bindpos[3] = {
-        (mjtNum) m->skin_bonebindpos[3*j],
-        (mjtNum) m->skin_bonebindpos[3*j+1],
-        (mjtNum) m->skin_bonebindpos[3*j+2]
-      };
-      mjtNum bindquat[4] = {
-        (mjtNum) m->skin_bonebindquat[4*j],
-        (mjtNum) m->skin_bonebindquat[4*j+1],
-        (mjtNum) m->skin_bonebindquat[4*j+2],
-        (mjtNum) m->skin_bonebindquat[4*j+3]
-      };
-
-      // compute rotation
-      int bodyid = m->skin_bonebodyid[j];
-      mjtNum quat[4], quatneg[4], rotate[9];
-      mju_negQuat(quatneg, bindquat);
-      mju_mulQuat(quat, d->xquat+4*bodyid, quatneg);
-      mju_quat2Mat(rotate, quat);
-
-      // compute translation
-      mjtNum translate[3];
-      mju_rotVecMat(translate, bindpos, rotate);
-      mju_sub3(translate, d->xpos+3*bodyid, translate);
-
-      // process all bone vertices
-      for (int k=m->skin_bonevertadr[j];
-           k<m->skin_bonevertadr[j]+m->skin_bonevertnum[j];
-           k++) {
-        // vertex id and weight
-        int vid = m->skin_bonevertid[k];
-        float vweight = m->skin_bonevertweight[k];
-
-        // get original position
-        mjtNum pos[3] = {
-          (mjtNum) m->skin_vert[3*(vertadr+vid)],
-          (mjtNum) m->skin_vert[3*(vertadr+vid)+1],
-          (mjtNum) m->skin_vert[3*(vertadr+vid)+2],
+    if (opt->skingroup[m->skin_group[i]]) {
+      // accumulate positions from all bones
+      for (int j=m->skin_boneadr[i];
+          j<m->skin_boneadr[i]+m->skin_bonenum[i];
+          j++) {
+        // get bind pose
+        mjtNum bindpos[3] = {
+          (mjtNum) m->skin_bonebindpos[3*j],
+          (mjtNum) m->skin_bonebindpos[3*j+1],
+          (mjtNum) m->skin_bonebindpos[3*j+2]
+        };
+        mjtNum bindquat[4] = {
+          (mjtNum) m->skin_bonebindquat[4*j],
+          (mjtNum) m->skin_bonebindquat[4*j+1],
+          (mjtNum) m->skin_bonebindquat[4*j+2],
+          (mjtNum) m->skin_bonebindquat[4*j+3]
         };
 
-        // transform
-        mjtNum pos1[3];
-        mju_rotVecMat(pos1, pos, rotate);
-        mju_addTo3(pos1, translate);
+        // compute rotation
+        int bodyid = m->skin_bonebodyid[j];
+        mjtNum quat[4], quatneg[4], rotate[9];
+        mju_negQuat(quatneg, bindquat);
+        mju_mulQuat(quat, d->xquat+4*bodyid, quatneg);
+        mju_quat2Mat(rotate, quat);
 
-        // accumulate position
-        scn->skinvert[3*(vertadr+vid)] += vweight*(float)pos1[0];
-        scn->skinvert[3*(vertadr+vid)+1] += vweight*(float)pos1[1];
-        scn->skinvert[3*(vertadr+vid)+2] += vweight*(float)pos1[2];
-      }
-    }
+        // compute translation
+        mjtNum translate[3];
+        mju_rotVecMat(translate, bindpos, rotate);
+        mju_sub3(translate, d->xpos+3*bodyid, translate);
 
-    // compute vertex normals from face normals
-    for (int k=faceadr; k<faceadr+facenum; k++) {
-      // get face vertex indices
-      int vid[3] = {
-        m->skin_face[3*k],
-        m->skin_face[3*k+1],
-        m->skin_face[3*k+2]
-      };
+        // process all bone vertices
+        for (int k=m->skin_bonevertadr[j];
+            k<m->skin_bonevertadr[j]+m->skin_bonevertnum[j];
+            k++) {
+          // vertex id and weight
+          int vid = m->skin_bonevertid[k];
+          float vweight = m->skin_bonevertweight[k];
 
-      // get triangle edges
-      mjtNum vec01[3], vec02[3];
-      for (int r=0; r<3; r++) {
-        vec01[r] = scn->skinvert[3*(vertadr+vid[1])+r] - scn->skinvert[3*(vertadr+vid[0])+r];
-        vec02[r] = scn->skinvert[3*(vertadr+vid[2])+r] - scn->skinvert[3*(vertadr+vid[0])+r];
-      }
+          // get original position
+          mjtNum pos[3] = {
+            (mjtNum) m->skin_vert[3*(vertadr+vid)],
+            (mjtNum) m->skin_vert[3*(vertadr+vid)+1],
+            (mjtNum) m->skin_vert[3*(vertadr+vid)+2],
+          };
 
-      // compute face normal
-      mjtNum nrm[3];
-      mju_cross(nrm, vec01, vec02);
+          // transform
+          mjtNum pos1[3];
+          mju_rotVecMat(pos1, pos, rotate);
+          mju_addTo3(pos1, translate);
 
-      // add normal to each vertex with weight = area
-      for (int r=0; r<3; r++) {
-        for (int t=0; t<3; t++) {
-          scn->skinnormal[3*(vertadr+vid[r])+t] += nrm[t];
+          // accumulate position
+          scn->skinvert[3*(vertadr+vid)] += vweight*(float)pos1[0];
+          scn->skinvert[3*(vertadr+vid)+1] += vweight*(float)pos1[1];
+          scn->skinvert[3*(vertadr+vid)+2] += vweight*(float)pos1[2];
         }
       }
-    }
 
-    // normalize normals
-    for (int k=vertadr; k<vertadr+vertnum; k++) {
-      float s = sqrtf(
-                  scn->skinnormal[3*k]*scn->skinnormal[3*k] +
-                  scn->skinnormal[3*k+1]*scn->skinnormal[3*k+1] +
-                  scn->skinnormal[3*k+2]*scn->skinnormal[3*k+2]
-                );
-      float scl = 1/mjMAX(mjMINVAL, s);
+      // compute vertex normals from face normals
+      for (int k=faceadr; k<faceadr+facenum; k++) {
+        // get face vertex indices
+        int vid[3] = {
+          m->skin_face[3*k],
+          m->skin_face[3*k+1],
+          m->skin_face[3*k+2]
+        };
 
-      scn->skinnormal[3*k] *= scl;
-      scn->skinnormal[3*k+1] *= scl;
-      scn->skinnormal[3*k+2] *= scl;
-    }
+        // get triangle edges
+        mjtNum vec01[3], vec02[3];
+        for (int r=0; r<3; r++) {
+          vec01[r] = scn->skinvert[3*(vertadr+vid[1])+r] - scn->skinvert[3*(vertadr+vid[0])+r];
+          vec02[r] = scn->skinvert[3*(vertadr+vid[2])+r] - scn->skinvert[3*(vertadr+vid[0])+r];
+        }
 
-    // inflate
-    if (m->skin_inflate[i]) {
-      float inflate = m->skin_inflate[i];
+        // compute face normal
+        mjtNum nrm[3];
+        mju_cross(nrm, vec01, vec02);
+
+        // add normal to each vertex with weight = area
+        for (int r=0; r<3; r++) {
+          for (int t=0; t<3; t++) {
+            scn->skinnormal[3*(vertadr+vid[r])+t] += nrm[t];
+          }
+        }
+      }
+
+      // normalize normals
       for (int k=vertadr; k<vertadr+vertnum; k++) {
-        scn->skinvert[3*k]   += inflate*scn->skinnormal[3*k];
-        scn->skinvert[3*k+1] += inflate*scn->skinnormal[3*k+1];
-        scn->skinvert[3*k+2] += inflate*scn->skinnormal[3*k+2];
+        float s = sqrtf(
+                    scn->skinnormal[3*k]*scn->skinnormal[3*k] +
+                    scn->skinnormal[3*k+1]*scn->skinnormal[3*k+1] +
+                    scn->skinnormal[3*k+2]*scn->skinnormal[3*k+2]
+                  );
+        float scl = 1/mjMAX(mjMINVAL, s);
+
+        scn->skinnormal[3*k] *= scl;
+        scn->skinnormal[3*k+1] *= scl;
+        scn->skinnormal[3*k+2] *= scl;
+      }
+
+      // inflate
+      if (m->skin_inflate[i]) {
+        float inflate = m->skin_inflate[i];
+        for (int k=vertadr; k<vertadr+vertnum; k++) {
+          scn->skinvert[3*k]   += inflate*scn->skinnormal[3*k];
+          scn->skinvert[3*k+1] += inflate*scn->skinnormal[3*k+1];
+          scn->skinvert[3*k+2] += inflate*scn->skinnormal[3*k+2];
+        }
       }
     }
   }
@@ -1875,6 +1998,181 @@ void mjv_updateScene(const mjModel* m, mjData* d, const mjvOption* opt,
 
   // update skins
   if (opt->flags[mjVIS_SKIN]) {
-    mjv_updateSkin(m, d, scn);
+    mjv_updateActiveSkin(m, d, scn, opt);
   }
 }
+
+
+
+//----------------------------------- catenary functions -------------------------------------------
+
+// returns hyperbolic cosine and optionally computes hyperbolic sine
+static inline mjtNum cosh_sinh(mjtNum x, mjtNum *sinh) {
+  mjtNum expx = mju_exp(x);
+  if (sinh) {
+    *sinh = 0.5 * (expx - 1/expx);
+  }
+  return 0.5 * (expx + 1/expx);
+}
+
+
+
+// returns intercept of the catenary equation
+static inline mjtNum catenary_intercept(mjtNum v, mjtNum h, mjtNum length) {
+  return 1/mju_sqrt(mju_sqrt(length*length - v*v)/h - 1);
+}
+
+
+
+// returns residual of catenary equation and optionally computes its gradient w.r.t b
+static inline mjtNum catenary_residual(mjtNum b, mjtNum intercept, mjtNum *grad) {
+  mjtNum a = 0.5 / b;
+  mjtNum sinh, cosh = cosh_sinh(a, &sinh);
+  if (grad) {
+    *grad = (a*cosh - sinh) * mju_pow(2*b*sinh - 1, -1.5);
+  }
+  return 1/mju_sqrt(2*b*sinh - 1) - intercept;
+}
+
+
+
+// convergence tolerance for catenary solver
+static const mjtNum tolerance = 1e-9;
+
+
+
+// solve trancendental catenary equation using change of variables proposed in
+//   https://math.stackexchange.com/a/1002996
+static inline mjtNum solve_catenary(mjtNum v, mjtNum h, mjtNum length) {
+  mjtNum intercept = catenary_intercept(v, h, length);
+
+  // initial guess using linear approximation to catenary_residual
+  mjtNum b = intercept / mju_sqrt(24);
+
+  // Newton steps to convergence (usually ~ 5 steps)
+  for (int i=0; i<50; i++) {
+    // get value and gradient
+    mjtNum grad;
+    mjtNum res = catenary_residual(b, intercept, &grad);
+
+    if (mju_abs(res) < tolerance) {
+      break;
+    }
+
+    // Newton step
+    mjtNum step = -res / grad;
+
+    // backtracking line-search is not essential but can reduce number of iterations
+    for (int j=0; j<10; j++) {
+      mjtNum new_res = catenary_residual(b + step, intercept, NULL);
+      if (mju_abs(new_res) < mju_abs(res)) {
+        break;
+      } else {
+        step *= 0.5;
+      }
+    }
+
+    // take step
+    b += step;
+  }
+
+  return b;
+}
+
+
+
+// points along catenary of given length between x0 and x1, returns number of points
+int mjv_catenary(const mjtNum x0[3], const mjtNum x1[3], const mjtNum gravity[3], mjtNum length,
+                 mjtNum catenary[3*mjNCATENARY]) {
+  mjtNum dist = mju_dist3(x0, x1);
+
+  // tendon is stretched longer than length: draw straight line
+  if (dist > length) {
+    // copy start and end points
+    mju_copy3(catenary+0, x0);
+    mju_copy3(catenary+3, x1);
+
+    return 2;
+  }
+
+  // tendon is shorter than length
+  else {
+    // normalized up vector
+    mjtNum up[3];
+    mju_scl3(up, gravity, -1);
+    mju_normalize3(up);
+
+    // x0 to x1
+    mjtNum x01[3];
+    mju_sub3(x01, x1, x0);
+
+    // make across orthonormal to up, points from x0 to x1
+    mjtNum across[3];
+    mju_copy3(across, x01);
+    mjtNum tmp[3];
+    mju_scl3(tmp, up, mju_dot3(up, across));
+    mju_subFrom3(across, tmp);
+    mjtNum norm = mju_normalize3(across);
+
+    // if across is numerically tiny, just set to 0
+    if (norm < mjMINVAL) {
+      mju_zero3(across);
+    }
+
+    // extents in the suspension plane
+    mjtNum h = mju_dot3(x01, across);  // horizontal suspension extent
+    mjtNum v = mju_dot3(x01, up);      // vertical height difference of x1 and x0
+
+    // near vertical tendon, use hanging bead approximation: 3 points
+    if (length > 100*h) {
+      // solve for location of bead hanging on tendon
+      mjtNum d_up = -0.5*(mju_sqrt(length*length - h*h) - v);  // down from x0
+      mjtNum d_across = h*d_up / (2*d_up - v);                 // across from x0
+
+      // start point
+      mju_copy3(catenary+0, x0);
+
+      // midpoint: bead location
+      mju_copy3(catenary+3, x0);
+      mju_addToScl3(catenary+3, up, d_up);
+      mju_addToScl3(catenary+3, across, d_across);
+
+      // end point
+      mju_copy3(catenary+6, x1);
+
+      return 3;
+    }
+
+    // compute catenary: mjNCATENARY points
+    else {
+      // b*h: scaled catenary flatness
+      mjtNum bh = solve_catenary(v, h, length) * h;
+
+      // horizontal and vertical offsets
+      mjtNum h_offset = -0.5 * (mju_log((length+v) / (length-v)) * bh - h);
+      mjtNum v_offset = -cosh_sinh(h_offset / bh, NULL) * bh;
+
+      // start point
+      mju_copy3(catenary+0, x0);
+
+      // hanging points
+      for (int i=1; i<mjNCATENARY-1; i++) {
+        // linearly spaced horizontal offset
+        mjtNum horizontal = i*h/mjNCATENARY;
+        mju_addScl3(catenary+3*i, x0, across, horizontal);
+
+        // vertical offset, evaluate catenary values
+        mjtNum vertical = bh * cosh_sinh((horizontal - h_offset) / bh, NULL) + v_offset;
+        mju_addToScl3(catenary+3*i, up, vertical);
+      }
+
+      // end point
+      mju_copy3(catenary+3*(mjNCATENARY-1), x1);
+
+      return mjNCATENARY;
+    }
+  }
+
+  return 0;  // SHOULD NOT OCCUR
+}
+
