@@ -62,11 +62,15 @@ mjCMesh::mjCMesh(mjCModel* _model, mjCDef* _def) {
   usernormal.clear();
   usertexcoord.clear();
   userface.clear();
+  useredge.clear();
 
   // clear internal variables
-  mjuu_setvec(pos, 0, 0, 0);
-  mjuu_setvec(quat, 1, 0, 0, 0);
-  mjuu_setvec(boxsz, 0, 0, 0);
+  mjuu_setvec(pos_surface, 0, 0, 0);
+  mjuu_setvec(pos_volume, 0, 0, 0);
+  mjuu_setvec(quat_surface, 1, 0, 0, 0);
+  mjuu_setvec(quat_volume, 1, 0, 0, 0);
+  mjuu_setvec(boxsz_surface, 0, 0, 0);
+  mjuu_setvec(boxsz_volume, 0, 0, 0);
   mjuu_setvec(aabb, 0, 0, 0);
   nvert = 0;
   nface = 0;
@@ -77,6 +81,12 @@ mjCMesh::mjCMesh(mjCModel* _model, mjCDef* _def) {
   face = NULL;
   graph = NULL;
   needhull = false;
+  validorientation = true;
+  validarea = true;
+  validvolume = true;
+  valideigenvalue = true;
+  validinequality = true;
+  processed = false;
 
   // reset to default if given
   if (_def) {
@@ -97,6 +107,7 @@ mjCMesh::~mjCMesh() {
   usernormal.clear();
   usertexcoord.clear();
   userface.clear();
+  useredge.clear();
 
   if (vert) mju_free(vert);
   if (normal) mju_free(normal);
@@ -200,6 +211,27 @@ void mjCMesh::Compile(const mjVFS* vfs) {
     nface = (int)userface.size()/3;
     face = (int*) mju_malloc(3*nface*sizeof(int));
     memcpy(face, userface.data(), 3*nface*sizeof(int));
+
+    // create half-edge structure (if mesh was in XML)
+    if (useredge.empty()) {
+      for (int i=0; i<nface; i++) {
+        int v0 = userface[3*i+0];
+        int v1 = userface[3*i+1];
+        int v2 = userface[3*i+2];
+        useredge.push_back(std::pair(v0, v1));
+        useredge.push_back(std::pair(v1, v2));
+        useredge.push_back(std::pair(v2, v0));
+      }
+    }
+  }
+
+  // check for inconsistent face orientations
+  if (!useredge.empty()) {
+    std::sort(useredge.begin(), useredge.end());
+    auto iterator = std::adjacent_find(useredge.begin(), useredge.end());
+    if (iterator != useredge.end()) {
+      validorientation = false;
+    }
   }
 
   // require vertices
@@ -224,25 +256,29 @@ void mjCMesh::Compile(const mjVFS* vfs) {
 
   // scale, center, orient, compute mass and inertia
   Process();
+  processed = true;
 }
 
 
 
 // get position
-void mjCMesh::GetPos(double* _pos) {
-  _pos[0] = pos[0];
-  _pos[1] = pos[1];
-  _pos[2] = pos[2];
+double* mjCMesh::GetPosPtr(mjtMeshType type) {
+  if (type==mjSHELL_MESH) {
+    return pos_surface;
+  } else {
+    return pos_volume;
+  }
 }
 
 
 
 // get orientation
-void mjCMesh::GetQuat(double* _quat) {
-  _quat[0] = quat[0];
-  _quat[1] = quat[1];
-  _quat[2] = quat[2];
-  _quat[3] = quat[3];
+double* mjCMesh::GetQuatPtr(mjtMeshType type) {
+  if (type==mjSHELL_MESH) {
+    return quat_surface;
+  } else {
+    return quat_volume;
+  }
 }
 
 
@@ -253,7 +289,8 @@ void mjCMesh::FitGeom(mjCGeom* geom, double* meshpos) {
 
   // use inertial box
   if (!model->fitaabb) {
-    // compute depending on type
+    // get inertia box type (shell or volume)
+    double* boxsz = GetInertiaBoxPtr(geom->typeinertia);
     switch (geom->type) {
     case mjGEOM_SPHERE:
       geom->size[0] = (boxsz[0] + boxsz[1] + boxsz[2])/3;
@@ -281,7 +318,7 @@ void mjCMesh::FitGeom(mjCGeom* geom, double* meshpos) {
     }
 
     // copy mesh pos into meshpos
-    mjuu_copyvec(meshpos, pos, 3);
+    mjuu_copyvec(meshpos, GetPosPtr(geom->typeinertia), 3);
   }
 
   // use AABB
@@ -503,7 +540,7 @@ void mjCMesh::LoadOBJ(const mjVFS* vfs) {
     std::stringstream msg;
     msg << "could not parse OBJ file '" << filename << "': \n"
         << objReader.Error();
-    throw mjCError(this, msg.str().c_str());
+    throw mjCError(this, "%s", msg.str().c_str());
   }
 
   auto attrib = objReader.GetAttrib();
@@ -525,18 +562,25 @@ void mjCMesh::LoadOBJ(const mjVFS* vfs) {
             this, "only tri or quad meshes are supported for OBJ (file '%s')",
             filename.c_str());
       }
+      // add face
       std::vector<std::array<tinyobj::index_t, 3>> faces;
-      std::array<tinyobj::index_t, 3> face1 = {
-          mesh.indices[index_in_mesh_indices],
-          mesh.indices[index_in_mesh_indices+1],
-          mesh.indices[index_in_mesh_indices+2]};
+      tinyobj::index_t v0 = mesh.indices[index_in_mesh_indices];
+      tinyobj::index_t v1 = mesh.indices[index_in_mesh_indices+1];
+      tinyobj::index_t v2 = mesh.indices[index_in_mesh_indices+2];
+      std::array<tinyobj::index_t, 3> face1 = {v0, v1, v2};
       faces.push_back(face1);
-      if (mesh.num_face_vertices[face] == 4) {  // add second triangle with 4th vertex
-        std::array<tinyobj::index_t, 3> face2 = {
-            mesh.indices[index_in_mesh_indices],
-            mesh.indices[index_in_mesh_indices+2],
-            mesh.indices[index_in_mesh_indices+3]};
+      // add edges
+      useredge.push_back(std::pair(v0.vertex_index, v1.vertex_index));
+      useredge.push_back(std::pair(v1.vertex_index, v2.vertex_index));
+      useredge.push_back(std::pair(v2.vertex_index, v0.vertex_index));
+      // handle quad: add second triangle with 4th vertex
+      if (mesh.num_face_vertices[face] == 4) {
+        tinyobj::index_t v3 = mesh.indices[index_in_mesh_indices+3];
+        std::array<tinyobj::index_t, 3> face2 = {v0, v2, v3};
         faces.push_back(face2);
+        useredge.push_back(std::pair(v0.vertex_index, v2.vertex_index));
+        useredge.push_back(std::pair(v2.vertex_index, v3.vertex_index));
+        useredge.push_back(std::pair(v3.vertex_index, v0.vertex_index));
       }
       for (const auto& face_indices : faces) {
         int index_of_first_vertex = uservert.size()/3;
@@ -587,36 +631,41 @@ void mjCMesh::LoadOBJ(const mjVFS* vfs) {
 // load STL binary mesh
 void mjCMesh::LoadSTL(const mjVFS* vfs) {
   bool righthand = (scale[0]*scale[1]*scale[2]>0);
-  int i, j;
 
   // make filename
   string filename = mjuu_makefullname(model->modelfiledir, model->meshdir, file);
 
   // get file data in buffer
   char* buffer = 0;
-  int buffer_sz = 0, flag_existing = 0;
+  int buffer_sz = 0;
+  bool own_buffer = false;
   if (vfs) {
     int id = mj_findFileVFS(vfs, filename.c_str());
     if (id>=0) {
       buffer = (char*)vfs->filedata[id];
       buffer_sz = vfs->filesize[id];
-      flag_existing = 1;
     }
   }
 
   // if not found in vfs, read from file
   if (!buffer) {
     buffer = (char*) mju_fileToMemory(filename.c_str(), &buffer_sz);
+    own_buffer = true;
   }
 
   // still not found
-  if (!buffer || !buffer_sz) {
+  if (!buffer) {
     throw mjCError(this, "could not open STL file '%s'", filename.c_str());
+  } else if (!buffer_sz) {
+    if (own_buffer) {
+      mju_free(buffer);
+    }
+    throw mjCError(this, "STL file '%s' is empty", filename.c_str());
   }
 
   // make sure there is enough data for header
   if (buffer_sz<84) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
 
@@ -626,7 +675,7 @@ void mjCMesh::LoadSTL(const mjVFS* vfs) {
   // get number of triangles, check bounds
   nface = *(unsigned int*)(buffer+80);
   if (nface<1 || nface>200000) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
 
@@ -637,7 +686,7 @@ void mjCMesh::LoadSTL(const mjVFS* vfs) {
 
   // check remaining buffer size
   if (nface*50 != buffer_sz-84) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
 
@@ -654,10 +703,30 @@ void mjCMesh::LoadSTL(const mjVFS* vfs) {
   vert = (float*) mju_malloc(9*nface*sizeof(float));
 
   // add vertices and faces, including repeated for now
-  for (i=0; i<nface; i++) {
-    for (j=0; j<3; j++) {
+  for (int i=0; i<nface; i++) {
+    for (int j=0; j<3; j++) {
       // get pointer to vertex coordiates
       float* v = (float*)(stl+50*i+12*(j+1));
+      for (int k=0; k < 3; k++) {
+        if (std::isnan(v[k]) || std::isinf(v[k])) {
+          if (own_buffer) {
+            mju_free(buffer);
+          }
+
+          throw mjCError(this, "STL file '%s' contains invalid vertices.",
+                         filename.c_str());
+        }
+        // check if vertex coordinates can be cast to an int safely
+        if (fabs(v[k])>pow(2, 30)) {
+          if (own_buffer) {
+            mju_free(buffer);
+          }
+
+          throw mjCError(this,
+                        "vertex coordinates in STL file '%s' exceed maximum bounds",
+                        filename.c_str());
+        }
+      }
 
       // add vertex address in face; change order if scale makes it lefthanded
       if (righthand || j==0) {
@@ -673,7 +742,7 @@ void mjCMesh::LoadSTL(const mjVFS* vfs) {
   }
 
   // free buffer if allocated here
-  if (!flag_existing) {
+  if (own_buffer) {
     mju_free(buffer);
   }
 
@@ -691,29 +760,35 @@ void mjCMesh::LoadMSH(const mjVFS* vfs) {
 
   // get file data in buffer
   char* buffer = 0;
-  int buffer_sz = 0, flag_existing = 0;
+  int buffer_sz = 0;
+  bool own_buffer = false;
   if (vfs) {
     int id = mj_findFileVFS(vfs, filename.c_str());
     if (id>=0) {
       buffer = (char*)vfs->filedata[id];
       buffer_sz = vfs->filesize[id];
-      flag_existing = 1;
     }
   }
 
   // if not found in vfs, read from file
   if (!buffer) {
     buffer = (char*) mju_fileToMemory(filename.c_str(), &buffer_sz);
+    own_buffer = true;
   }
 
   // still not found
-  if (!buffer || !buffer_sz) {
+  if (!buffer) {
     throw mjCError(this, "could not open MSH file '%s'", filename.c_str());
+  } else if (!buffer_sz) {
+    if (own_buffer) {
+      mju_free(buffer);
+    }
+    throw mjCError(this, "MSH file '%s' is empty", filename.c_str());
   }
 
   // make sure header is present
   if (buffer_sz<4*sizeof(int)) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
     throw mjCError(this, "missing header in MSH file '%s'", filename.c_str());
@@ -729,7 +804,7 @@ void mjCMesh::LoadMSH(const mjVFS* vfs) {
   if (nvert<4 || nface<0 || nnormal<0 || ntexcoord<0 ||
       (nnormal>0 && nnormal!=nvert) ||
       (ntexcoord>0 && ntexcoord!=nvert)) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
     throw mjCError(this, "invalid sizes in MSH file '%s'", filename.c_str());
@@ -738,7 +813,7 @@ void mjCMesh::LoadMSH(const mjVFS* vfs) {
   // check file size
   if (buffer_sz != 4*sizeof(int) + 3*nvert*sizeof(float) + 3*nnormal*sizeof(float) +
       2*ntexcoord*sizeof(float) + 3*nface*sizeof(int)) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
     throw mjCError(this, "unexpected file size in MSH file '%s'", filename.c_str());
@@ -777,7 +852,7 @@ void mjCMesh::LoadMSH(const mjVFS* vfs) {
   }
 
   // free buffer if allocated here
-  if (!flag_existing) {
+  if (own_buffer) {
     mju_free(buffer);
   }
 }
@@ -820,252 +895,304 @@ static double _areaNrmCen(double* normal, double* center,
 
 
 // apply transformations
-void mjCMesh::Process(void) {
-  double CoM[3] = {0, 0, 0};
-  double facecen[3] = {0, 0, 0};
-  double area = 0;
-  double inert[6] = {0, 0, 0, 0, 0, 0};
-  double volume;
+void mjCMesh::Process() {
+  for ( const auto type : { mjtMeshType::mjVOLUME_MESH, mjtMeshType::mjSHELL_MESH } ) {
+    double CoM[3] = {0, 0, 0};
+    double facecen[3] = {0, 0, 0};
+    double area = 0;
+    double inert[6] = {0, 0, 0, 0, 0, 0};
 
-  int i, j;
-  double nrm[3];
-  double cen[3];
+    int i, j;
+    double nrm[3];
+    double cen[3];
 
-  // translate
-  if (refpos[0]!=0 || refpos[1]!=0 || refpos[2]!=0) {
-    // prepare translation
-    float rp[3] = {(float)refpos[0], (float)refpos[1], (float)refpos[2]};
+    if (type==mjVOLUME_MESH) {
+      // translate
+      if (refpos[0]!=0 || refpos[1]!=0 || refpos[2]!=0) {
+        // prepare translation
+        float rp[3] = {(float)refpos[0], (float)refpos[1], (float)refpos[2]};
 
-    // process vertices
-    for (i=0; i<nvert; i++) {
-      // positions
-      vert[3*i] -= rp[0];
-      vert[3*i+1] -= rp[1];
-      vert[3*i+2] -= rp[2];
+        // process vertices
+        for (i=0; i<nvert; i++) {
+          // positions
+          vert[3*i] -= rp[0];
+          vert[3*i+1] -= rp[1];
+          vert[3*i+2] -= rp[2];
 
-      // normals not affected by translation
-    }
-  }
+          // normals not affected by translation
+        }
+      }
 
-  // rotate
-  if (refquat[0]!=1 || refquat[1]!=0 || refquat[2]!=0 || refquat[3]!=0) {
-    // prepare rotation
-    mjtNum quat[4] = {refquat[0], refquat[1], refquat[2], refquat[3]};
-    mjtNum mat[9];
-    mju_normalize4(quat);
-    mju_quat2Mat(mat, quat);
+      // rotate
+      if (refquat[0]!=1 || refquat[1]!=0 || refquat[2]!=0 || refquat[3]!=0) {
+        // prepare rotation
+        mjtNum quat[4] = {refquat[0], refquat[1], refquat[2], refquat[3]};
+        mjtNum mat[9];
+        mju_normalize4(quat);
+        mju_quat2Mat(mat, quat);
 
-    // process vertices
-    for (i=0; i<nvert; i++) {
-      // positions
-      mjtNum p1[3], p0[3] = {vert[3*i], vert[3*i+1], vert[3*i+2]};
-      mju_rotVecMatT(p1, p0, mat);
-      vert[3*i] = (float) p1[0];
-      vert[3*i+1] = (float) p1[1];
-      vert[3*i+2] = (float) p1[2];
+        // process vertices
+        for (i=0; i<nvert; i++) {
+          // positions
+          mjtNum p1[3], p0[3] = {vert[3*i], vert[3*i+1], vert[3*i+2]};
+          mju_rotVecMatT(p1, p0, mat);
+          vert[3*i] = (float) p1[0];
+          vert[3*i+1] = (float) p1[1];
+          vert[3*i+2] = (float) p1[2];
 
-      // nromals
-      mjtNum n1[3], n0[3] = {normal[3*i], normal[3*i+1], normal[3*i+2]};
-      mju_rotVecMatT(n1, n0, mat);
-      normal[3*i] = (float) n1[0];
-      normal[3*i+1] = (float) n1[1];
-      normal[3*i+2] = (float) n1[2];
-    }
-  }
+          // normals
+          mjtNum n1[3], n0[3] = {normal[3*i], normal[3*i+1], normal[3*i+2]};
+          mju_rotVecMatT(n1, n0, mat);
+          normal[3*i] = (float) n1[0];
+          normal[3*i+1] = (float) n1[1];
+          normal[3*i+2] = (float) n1[2];
+        }
+      }
 
-  // scale
-  if (scale[0]!=1 || scale[1]!=1 || scale[2]!=1) {
-    for (i=0; i<nvert; i++) {
-      // positions
-      vert[3*i] *= scale[0];
-      vert[3*i+1] *= scale[1];
-      vert[3*i+2] *= scale[2];
+      // scale
+      if (scale[0]!=1 || scale[1]!=1 || scale[2]!=1) {
+        for (i=0; i<nvert; i++) {
+          // positions
+          vert[3*i] *= scale[0];
+          vert[3*i+1] *= scale[1];
+          vert[3*i+2] *= scale[2];
 
-      // normals
-      normal[3*i] *= scale[0];
-      normal[3*i+1] *= scale[1];
-      normal[3*i+2] *= scale[2];
-    }
-  }
+          // normals
+          normal[3*i] *= scale[0];
+          normal[3*i+1] *= scale[1];
+          normal[3*i+2] *= scale[2];
+        }
+      }
 
-  // normalize normals
-  for (i=0; i<nvert; i++) {
-    // compute length
-    float len = normal[3*i]*normal[3*i] + normal[3*i+1]*normal[3*i+1] + normal[3*i+2]*normal[3*i+2];
+      // normalize normals
+      for (i=0; i<nvert; i++) {
+        // compute length
+        float len = normal[3*i]*normal[3*i] + normal[3*i+1]*normal[3*i+1] + normal[3*i+2]*normal[3*i+2];
 
-    // rescale
-    if (len>mjMINVAL) {
-      float scl = 1/sqrtf(len);
-      normal[3*i] *= scl;
-      normal[3*i+1] *= scl;
-      normal[3*i+2] *= scl;
-    } else {
-      normal[3*i] = 0;
-      normal[3*i+1] = 0;
-      normal[3*i+2] = 1;
-    }
-  }
+        // rescale
+        if (len>mjMINVAL) {
+          float scl = 1/sqrtf(len);
+          normal[3*i] *= scl;
+          normal[3*i+1] *= scl;
+          normal[3*i+2] *= scl;
+        } else {
+          normal[3*i] = 0;
+          normal[3*i+1] = 0;
+          normal[3*i+2] = 1;
+        }
+      }
 
-  // find centroid of faces
-  for (i=0; i<nface; i++) {
-    // check vertex indices
-    for (j=0; j<3; j++) {
-      if (face[3*i+j]<0 || face[3*i+j]>=nvert) {
-        throw mjCError(this, "vertex index out of range in %s (index = %d)", name.c_str(), i);
+      // find centroid of faces
+      for (i=0; i<nface; i++) {
+        // check vertex indices
+        for (j=0; j<3; j++) {
+          if (face[3*i+j]<0 || face[3*i+j]>=nvert) {
+            throw mjCError(this, "vertex index out of range in %s (index = %d)", name.c_str(), i);
+          }
+        }
+
+        // get area and center
+        double a = _areaNrmCen(nrm, cen, vert+3*face[3*i], vert+3*face[3*i+1], vert+3*face[3*i+2]);
+
+        // accumulate
+        for (j=0; j<3; j++) {
+          facecen[j] += a*cen[j];
+        }
+        area += a;
+      }
+
+      // require positive area
+      if (area < mjMINVAL) {
+        validarea = false;
+        return;
+      }
+
+      // finalize centroid of faces
+      for (j=0; j<3; j++) {
+        facecen[j] /= area;
       }
     }
 
-    // get area and center
-    double a = _areaNrmCen(nrm, cen, vert+3*face[3*i], vert+3*face[3*i+1], vert+3*face[3*i+2]);
+    // compute CoM and volume from pyramid volumes
+    GetVolumeRef(type) = 0;
+    for (i=0; i<nface; i++) {
+      // get area, normal and center
+      double a = _areaNrmCen(nrm, cen, vert+3*face[3*i], vert+3*face[3*i+1], vert+3*face[3*i+2]);
 
-    // accumulate
-    for (j=0; j<3; j++) {
-      facecen[j] += a*cen[j];
-    }
-    area += a;
-  }
+      // compute and add volume
+      const double vec[3] = {cen[0]-facecen[0], cen[1]-facecen[1], cen[2]-facecen[2]};
+      double vol = type==mjSHELL_MESH ? a : mjuu_dot3(vec, nrm) * a / 3;
 
-  // require positive area
-  if (area < mjMINVAL) {
-    throw mjCError(this, "mesh surface area is too small: %s", name.c_str());
-  }
+      // if legacy computation requested, then always positive
+      if (!model->exactmeshinertia) {
+        vol = fabs(vol);
+      }
 
-  // finalize centroid of faces
-  for (j=0; j<3; j++) {
-    facecen[j] /= area;
-  }
-
-  // compute CoM and volume from pyramid volumes
-  volume = 0;
-  for (i=0; i<nface; i++) {
-    // get area, normal and center
-    double a = _areaNrmCen(nrm, cen, vert+3*face[3*i], vert+3*face[3*i+1], vert+3*face[3*i+2]);
-
-    // compute and add volume
-    const double vec[3] = {cen[0]-facecen[0], cen[1]-facecen[1], cen[2]-facecen[2]};
-    double vol = fabs(mjuu_dot3(vec, nrm)) * a / 3;
-    volume += vol;
-
-    // add pyramid com
-    for (j=0; j<3; j++) {
-      CoM[j] += vol*(cen[j]*3.0/4.0 + facecen[j]/4.0);
-    }
-  }
-
-  // require positive volume
-  if (volume < mjMINVAL) {
-    throw mjCError(this, "mesh volume is too small: %s", name.c_str());
-  }
-
-  // finalize CoM, save as mesh center
-  for (j=0; j<3; j++) {
-    CoM[j] /= volume;
-  }
-  mjuu_copyvec(pos, CoM, 3);
-
-  // re-center mesh at CoM
-  for (i=0; i<nvert; i++) {
-    for (j=0; j<3; j++) {
-      vert[3*i+j] -= CoM[j];
-    }
-  }
-
-  // accumulate products of inertia, recompute volume
-  const int k[6][2] = {{0, 0}, {1, 1}, {2, 2}, {0, 1}, {0, 2}, {1, 2}};
-  double P[6] = {0, 0, 0, 0, 0, 0};
-  volume = 0;
-  for (i=0; i<nface; i++) {
-    float* D = vert+3*face[3*i];
-    float* E = vert+3*face[3*i+1];
-    float* F = vert+3*face[3*i+2];
-
-    // get area, normal and center; update volume
-    double a = _areaNrmCen(nrm, cen, D, E, F);
-    double vol = fabs(mjuu_dot3(cen, nrm)) * a / 3;
-    volume += vol;
-
-    // apply formula, accumulate
-    for (j=0; j<6; j++) {
-      P[j] += def->geom.density*vol/20 * (
-                2*(D[k[j][0]] * D[k[j][1]] +
-                   E[k[j][0]] * E[k[j][1]] +
-                   F[k[j][0]] * F[k[j][1]]) +
-                D[k[j][0]] * E[k[j][1]]  +  D[k[j][1]] * E[k[j][0]] +
-                D[k[j][0]] * F[k[j][1]]  +  D[k[j][1]] * F[k[j][0]] +
-                E[k[j][0]] * F[k[j][1]]  +  E[k[j][1]] * F[k[j][0]]);
-    }
-  }
-
-  // convert from products of inertia to moments of inertia
-  inert[0] = P[1] + P[2];
-  inert[1] = P[0] + P[2];
-  inert[2] = P[0] + P[1];
-  inert[3] = -P[3];
-  inert[4] = -P[4];
-  inert[5] = -P[5];
-
-  // get quaternion and diagonal inertia
-  mjtNum eigval[3], eigvec[9], quattmp[4];
-  mjtNum full[9] = {
-    inert[0], inert[3], inert[4],
-    inert[3], inert[1], inert[5],
-    inert[4], inert[5], inert[2]
-  };
-  mju_eig3(eigval, eigvec, quattmp, full);
-
-  // check eigval - SHOULD NOT OCCUR
-  if (eigval[2]<mjEPS) {
-    throw mjCError(this, "eigenvalue of mesh inertia too small: %s", name.c_str());
-  }
-  if (eigval[0] + eigval[1] < eigval[2] ||
-      eigval[0] + eigval[2] < eigval[1] ||
-      eigval[1] + eigval[2] < eigval[0]) {
-    throw mjCError(this,
-                   "eigenvalues of mesh inertia violate A + B >= C condition: %s", name.c_str());
-  }
-
-  // compute sizes of equivalent inertia box
-  double mass = volume * def->geom.density;
-  boxsz[0] = sqrt(6*(eigval[1]+eigval[2]-eigval[0])/mass)/2;
-  boxsz[1] = sqrt(6*(eigval[0]+eigval[2]-eigval[1])/mass)/2;
-  boxsz[2] = sqrt(6*(eigval[0]+eigval[1]-eigval[2])/mass)/2;
-
-  // copy quat
-  for (j=0; j<4; j++) {
-    quat[j] = quattmp[j];
-  }
-
-  // rotate vertices and normals into axis-aligned frame
-  double neg[4] = {quattmp[0], -quattmp[1], -quattmp[2], -quattmp[3]};
-  double mat[9];
-  mjuu_quat2mat(mat, neg);
-  for (i=0; i<nvert; i++) {
-    // vertices
-    const double vec[3] = {vert[3*i], vert[3*i+1], vert[3*i+2]};
-    double res[3];
-    mjuu_mulvecmat(res, vec, mat);
-    for (j=0; j<3; j++) {
-      vert[3*i+j] = (float) res[j];
+      // add pyramid com
+      GetVolumeRef(type) += vol;
+      for (j=0; j<3; j++) {
+        CoM[j] += vol*(cen[j]*3.0/4.0 + facecen[j]/4.0);
+      }
     }
 
-    // normals
-    const double nrm[3] = {normal[3*i], normal[3*i+1], normal[3*i+2]};
-    mjuu_mulvecmat(res, nrm, mat);
-    for (j=0; j<3; j++) {
-      normal[3*i+j] = (float) res[j];
+    // require positive volume
+    if (GetVolumeRef(type) < mjMINVAL) {
+      validvolume = false;
+      return;
     }
-  }
 
-  // compute axis-aligned bounding box
-  for (i=0; i<nvert; i++) {
-    float* v = vert+3*i;
+    // finalize CoM, save as mesh center
     for (j=0; j<3; j++) {
-      aabb[j] = mjMAX(aabb[j], fabs(v[j]));
+      CoM[j] /= GetVolumeRef(type);
+    }
+    mjuu_copyvec(GetPosPtr(type), CoM, 3);
+
+    // re-center mesh at CoM
+    if (type==mjVOLUME_MESH) {
+      for (i=0; i<nvert; i++) {
+        for (j=0; j<3; j++) {
+          vert[3*i+j] -= CoM[j];
+        }
+      }
+    }
+
+    // accumulate products of inertia, recompute volume
+    const int k[6][2] = {{0, 0}, {1, 1}, {2, 2}, {0, 1}, {0, 2}, {1, 2}};
+    double P[6] = {0, 0, 0, 0, 0, 0};
+    GetVolumeRef(type) = 0;
+    for (i=0; i<nface; i++) {
+      float* D = vert+3*face[3*i];
+      float* E = vert+3*face[3*i+1];
+      float* F = vert+3*face[3*i+2];
+
+      // get area, normal and center; update volume
+      double a = _areaNrmCen(nrm, cen, D, E, F);
+      double vol = type==mjSHELL_MESH ? a : mjuu_dot3(cen, nrm) * a / 3;
+
+      // if legacy computation requested, then always positive
+      if (!model->exactmeshinertia) {
+        vol = fabs(vol);
+      }
+
+      // apply formula, accumulate
+      GetVolumeRef(type) += vol;
+      for (j=0; j<6; j++) {
+        P[j] += def->geom.density*vol /
+                  (type==mjSHELL_MESH ? 12 : 20) * (
+                  2*(D[k[j][0]] * D[k[j][1]] +
+                    E[k[j][0]] * E[k[j][1]] +
+                    F[k[j][0]] * F[k[j][1]]) +
+                  D[k[j][0]] * E[k[j][1]]  +  D[k[j][1]] * E[k[j][0]] +
+                  D[k[j][0]] * F[k[j][1]]  +  D[k[j][1]] * F[k[j][0]] +
+                  E[k[j][0]] * F[k[j][1]]  +  E[k[j][1]] * F[k[j][0]]);
+      }
+    }
+
+    // convert from products of inertia to moments of inertia
+    inert[0] = P[1] + P[2];
+    inert[1] = P[0] + P[2];
+    inert[2] = P[0] + P[1];
+    inert[3] = -P[3];
+    inert[4] = -P[4];
+    inert[5] = -P[5];
+
+    // get quaternion and diagonal inertia
+    mjtNum eigval[3], eigvec[9], quattmp[4];
+    mjtNum full[9] = {
+      inert[0], inert[3], inert[4],
+      inert[3], inert[1], inert[5],
+      inert[4], inert[5], inert[2]
+    };
+    mju_eig3(eigval, eigvec, quattmp, full);
+
+    // check eigval - SHOULD NOT OCCUR
+    if (eigval[2]<=0) {
+      valideigenvalue = false;
+      return;
+    }
+    if (eigval[0] + eigval[1] < eigval[2] ||
+        eigval[0] + eigval[2] < eigval[1] ||
+        eigval[1] + eigval[2] < eigval[0]) {
+      validinequality = false;
+      return;
+    }
+
+    // compute sizes of equivalent inertia box
+    double mass = GetVolumeRef(type) * def->geom.density;
+    double* boxsz = GetInertiaBoxPtr(type);
+    boxsz[0] = sqrt(6*(eigval[1]+eigval[2]-eigval[0])/mass)/2;
+    boxsz[1] = sqrt(6*(eigval[0]+eigval[2]-eigval[1])/mass)/2;
+    boxsz[2] = sqrt(6*(eigval[0]+eigval[1]-eigval[2])/mass)/2;
+
+    // copy quat
+    for (j=0; j<4; j++) {
+      GetQuatPtr(type)[j] = quattmp[j];
+    }
+
+    // rotate vertices and normals into axis-aligned frame
+    if (type==mjVOLUME_MESH) {
+      double neg[4] = {quattmp[0], -quattmp[1], -quattmp[2], -quattmp[3]};
+      double mat[9];
+      mjuu_quat2mat(mat, neg);
+      for (i=0; i<nvert; i++) {
+        // vertices
+        const double vec[3] = {vert[3*i], vert[3*i+1], vert[3*i+2]};
+        double res[3];
+        mjuu_mulvecmat(res, vec, mat);
+        for (j=0; j<3; j++) {
+          vert[3*i+j] = (float) res[j];
+        }
+
+        // normals
+        const double nrm[3] = {normal[3*i], normal[3*i+1], normal[3*i+2]};
+        mjuu_mulvecmat(res, nrm, mat);
+        for (j=0; j<3; j++) {
+          normal[3*i+j] = (float) res[j];
+        }
+      }
+
+      // compute axis-aligned bounding box
+      for (i=0; i<nvert; i++) {
+        float* v = vert+3*i;
+        for (j=0; j<3; j++) {
+          aabb[j] = mjMAX(aabb[j], fabs(v[j]));
+        }
+      }
     }
   }
 }
 
 
+// check that the mesh is valid
+void mjCMesh::CheckMesh() {
+  if (!processed) {
+    return;
+  }
+  if (!validorientation)
+    throw mjCError(this, "faces have inconsistent orientation: %s", name.c_str());
+  if (!validarea)
+    throw mjCError(this, "mesh surface area is too small: %s", name.c_str());
+  if (!validvolume)
+    throw mjCError(this, "mesh volume is too small: %s", name.c_str());
+  if (!valideigenvalue)
+    throw mjCError(this, "eigenvalue of mesh inertia must be positive: %s", name.c_str());
+  if (!validinequality)
+    throw mjCError(this, "eigenvalues of mesh inertia violate A + B >= C: %s", name.c_str());
+}
+
+
+// get inertia pointer
+double* mjCMesh::GetInertiaBoxPtr(mjtMeshType type) {
+  CheckMesh();
+  return type==mjSHELL_MESH ? boxsz_surface : boxsz_volume;
+}
+
+
+double& mjCMesh::GetVolumeRef(mjtMeshType type) {
+  CheckMesh();
+  return type==mjSHELL_MESH ? surface : volume;
+}
+
 
 // make graph describing convex hull
 void mjCMesh::MakeGraph(void) {
@@ -1408,6 +1535,7 @@ mjCSkin::mjCSkin(mjCModel* _model) {
   rgba[0] = rgba[1] = rgba[2] = 0.5f;
   rgba[3] = 1.0f;
   inflate = 0;
+  group = 0;
 
   vert.clear();
   texcoord.clear();
@@ -1530,7 +1658,6 @@ void mjCSkin::Compile(const mjVFS* vfs) {
       throw mjCError(this, "unkown material '%s' in skin", material.c_str());
   }
 
-
   // set total vertex weights to 0
   vector<float> vw;
   size_t nvert = vert.size()/3;
@@ -1598,29 +1725,35 @@ void mjCSkin::LoadSKN(const mjVFS* vfs) {
 
   // get file data in buffer
   char* buffer = NULL;
-  int buffer_sz = 0, flag_existing = 0;
+  int buffer_sz = 0;
+  bool own_buffer = false;
   if (vfs) {
     int id = mj_findFileVFS(vfs, filename.c_str());
     if (id>=0) {
       buffer = (char*)vfs->filedata[id];
       buffer_sz = vfs->filesize[id];
-      flag_existing = 1;
     }
   }
 
   // if not found in vfs, read from file
   if (!buffer) {
     buffer = (char*) mju_fileToMemory(filename.c_str(), &buffer_sz);
+    own_buffer = true;
   }
 
   // still not found
-  if (!buffer || !buffer_sz) {
+  if (!buffer) {
     throw mjCError(this, "could not open SKN file '%s'", filename.c_str());
+  } else if (!buffer_sz) {
+    if (own_buffer) {
+      mju_free(buffer);
+    }
+    throw mjCError(this, "SKN file '%s' is empty", filename.c_str());
   }
 
   // make sure header is present
   if (buffer_sz<16) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
     throw mjCError(this, "missing header in SKN file '%s'", filename.c_str());
@@ -1634,7 +1767,7 @@ void mjCSkin::LoadSKN(const mjVFS* vfs) {
 
   // negative sizes not allowed
   if (nvert<0 || ntexcoord<0 || nface<0 || nbone<0) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
     throw mjCError(this, "negative size in header of SKN file '%s'", filename.c_str());
@@ -1642,7 +1775,7 @@ void mjCSkin::LoadSKN(const mjVFS* vfs) {
 
   // make sure we have data for vert, texcoord, face
   if (buffer_sz < 16 + 12*nvert + 8*ntexcoord + 12*nface) {
-    if (!flag_existing) {
+    if (own_buffer) {
       mju_free(buffer);
     }
     throw mjCError(this, "insufficient data in SKN file '%s'", filename.c_str());
@@ -1684,7 +1817,7 @@ void mjCSkin::LoadSKN(const mjVFS* vfs) {
   for (int i=0; i<nbone; i++) {
     // check size
     if (buffer_sz/4-4-cnt < 18) {
-      if (!flag_existing) {
+      if (own_buffer) {
         mju_free(buffer);
       }
       throw mjCError(this, "insufficient data in SKN file '%s', bone %d", filename.c_str(), i);
@@ -1711,7 +1844,7 @@ void mjCSkin::LoadSKN(const mjVFS* vfs) {
 
     // check for negative
     if (vcount<1) {
-      if (!flag_existing) {
+      if (own_buffer) {
         mju_free(buffer);
       }
       throw mjCError(this, "vertex count must be positive in SKN file '%s', bone %d",
@@ -1720,7 +1853,7 @@ void mjCSkin::LoadSKN(const mjVFS* vfs) {
 
     // check size
     if (buffer_sz/4-4-cnt < 2*vcount) {
-      if (!flag_existing) {
+      if (own_buffer) {
         mju_free(buffer);
       }
       throw mjCError(this, "insufficient vertex data in SKN file '%s', bone %d",
@@ -1739,7 +1872,7 @@ void mjCSkin::LoadSKN(const mjVFS* vfs) {
   }
 
   // free buffer if allocated here
-  if (!flag_existing) {
+  if (own_buffer) {
     mju_free(buffer);
   }
 
