@@ -17,6 +17,7 @@
 #include "src/engine/engine_io.h"
 
 #include <array>
+#include <climits>
 #include <cstring>
 #include <string>
 
@@ -24,6 +25,7 @@
 #include <gtest/gtest.h>
 #include <absl/strings/str_format.h>
 #include <mujoco/mjxmacro.h>
+#include "src/engine/engine_util_errmem.h"
 #include "test/fixture.h"
 
 namespace mujoco {
@@ -70,12 +72,17 @@ TEST_F(EngineIoTest, MakeDataFromPartialModel) {
   ASSERT_THAT(data_from_partial, NotNull());
 
   EXPECT_EQ(data_from_partial->nbuffer, data_from_model->nbuffer);
-  int nbuffer = data_from_partial->nbuffer;
   // If there are no mocap bodies and qpos0 is all zero, mjData should be the
   // same whether it was made from the full model or the partial model.
-  EXPECT_EQ(
-      std::memcmp(data_from_partial->buffer, data_from_model->buffer, nbuffer),
-      0) << "mjData content differs";
+  {
+    MJDATA_POINTERS_PREAMBLE((&partial_model))
+    #define X(type, name, nr, nc)                                              \
+        EXPECT_EQ(std::memcmp(data_from_partial->name, data_from_model->name,  \
+                              sizeof(type)*(partial_model.nr)*(nc)),           \
+                  0) << "mjData::" #name " differs";
+    MJDATA_POINTERS
+    #undef X
+  }
 
   mj_deleteData(data_from_model);
   mj_deleteData(data_from_partial);
@@ -158,15 +165,42 @@ TEST_F(EngineIoTest, CopyDataWithPartialModel) {
 
   EXPECT_EQ(copy->nbuffer, data->nbuffer);
   EXPECT_EQ(copy->qpos[0], 1);
-  int nbuffer = copy->nbuffer;
-  EXPECT_EQ(
-      std::memcmp(copy->buffer, data->buffer, nbuffer),
-      0) << "mjData content differs";
+  {
+    MJDATA_POINTERS_PREAMBLE((&partial_model))
+    #define X(type, name, nr, nc)                                     \
+        EXPECT_EQ(std::memcmp(copy->name, data->name,                 \
+                              sizeof(type)*(partial_model.nr)*(nc)),  \
+                  0) << "mjData::" #name " differs";
+    MJDATA_POINTERS
+    #undef X
+  }
 
   mj_deleteData(data);
   mj_deleteData(copy);
 }
 
+TEST_F(EngineIoTest, MakeDataReturnsNullOnFailure) {
+  constexpr char xml[] = "<mujoco/>";
+
+  std::array<char, 1024> error;
+  mjModel* model = LoadModelFromString(xml, error.data(), error.size());
+  ASSERT_THAT(model, NotNull()) << "Failed to load model: " << error.data();
+
+  // fail mj_makeData intentionally with a bad size
+  model->nbody = -1;
+  static bool warning;
+  warning = false;
+  mju_user_warning = [](const char* error) {
+    warning = true;
+  };
+  mjData* data = mj_makeData(model);
+  EXPECT_THAT(data, IsNull());
+  EXPECT_TRUE(warning) << "Expecting warning to be triggered.";
+
+  mj_deleteData(data);
+  mj_deleteModel(model);
+}
+
 using ValidateReferencesTest = MujocoTest;
 
 TEST_F(ValidateReferencesTest, BodyReferences) {
@@ -223,6 +257,32 @@ TEST_F(ValidateReferencesTest, AddressRange) {
   mj_deleteModel(model);
 }
 
+TEST_F(ValidateReferencesTest, AddressRangeNegativeNum) {
+  static const char xml[] = R"(
+  <mujoco>
+    <worldbody>
+      <body>
+        <joint/>
+        <joint/>
+        <geom size="1"/>
+      </body>
+    </worldbody>
+  </mujoco>
+  )";
+
+  std::array<char, 1024> error;
+  mjModel* model = LoadModelFromString(xml, error.data(), error.size());
+  ASSERT_THAT(model, NotNull()) << "Failed to load model: " << error.data();
+
+  EXPECT_THAT(mj_validateReferences(model), IsNull());
+
+  // jntadr + jntnum is within safe range, but jntnum is negative.
+  model->body_jntadr[1] += 5;
+  model->body_jntnum[1] -= 5;
+  EXPECT_THAT(mj_validateReferences(model), HasSubstr("body_jntnum"));
+  mj_deleteModel(model);
+}
+
 TEST_F(ValidateReferencesTest, GeomCondim) {
   static const char xml[] = R"(
   <mujoco>
@@ -489,7 +549,6 @@ TEST_F(ValidateReferencesTest, EqualityConstraints) {
     <equality>
       <connect anchor="0 0 0" body1="body1" />
       <weld body1="body1" body2="body2" />
-      <distance geom1="geom1" geom2="geom2" />
       <joint joint1="joint1"/>
       <joint joint1="joint1" joint2="joint2"/>
       <tendon tendon1="tendon1"/>
@@ -530,18 +589,6 @@ TEST_F(ValidateReferencesTest, EqualityConstraints) {
   EXPECT_THAT(mj_validateReferences(model), HasSubstr("eq_obj2id"));
   model->eq_obj2id[1] = model->nbody - 1;
 
-  // distance constraint
-  model->eq_obj1id[2] = -1;
-  EXPECT_THAT(mj_validateReferences(model), HasSubstr("eq_obj1id"));
-  model->eq_obj1id[2] = model->ngeom;
-  EXPECT_THAT(mj_validateReferences(model), HasSubstr("eq_obj1id"));
-  model->eq_obj1id[2] = 1;
-
-  model->eq_obj2id[2] = -1;
-  EXPECT_THAT(mj_validateReferences(model), HasSubstr("eq_obj2id"));
-  model->eq_obj2id[2] = model->ngeom;
-  EXPECT_THAT(mj_validateReferences(model), HasSubstr("eq_obj2id"));
-  model->eq_obj2id[2] = model->ngeom - 1;
   mj_deleteModel(model);
 }
 
